// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package cmd

import (
	"fmt"
	"os"
	"syscall"
	"time"

	goCmd "github.com/go-cmd/cmd"
)

type ErrTimeout struct {
	Cmd string
}

func (e ErrTimeout) Error() string {
	return fmt.Sprint("timed out running command: ", e.Cmd)
}

type ErrCmdExitedNonZero struct {
	Ret int
}

func (e ErrCmdExitedNonZero) Error() string {
	return fmt.Sprint("command exited with non-zero return code: ", e.Ret)
}

type ErrCmdExited struct{}

func (e ErrCmdExited) Error() string {
	return "command exited"
}

type ErrUnhandledStatus struct {
	Msg string
}

func (e ErrUnhandledStatus) Error() string {
	return fmt.Sprint("unhandled status received: ", e.Msg)
}

var (
	// for mocking in tests
	execCommand = goCmd.NewCmdOptions
	// give 100ms for process to start
	startTimerDuration = 100 * time.Millisecond
	upCheck            = isUp
)

type Streams struct {
	// TODO: wrap goCmd.Status
	Status <-chan goCmd.Status
	Stdout <-chan string
	Stderr <-chan string
}

type Commander interface {
	Status() goCmd.Status
	Start(env []string) (err error)
	Stop() error
	Streams() *Streams
}

type Command struct {
	cmd     *goCmd.Cmd
	streams *Streams
}

func NewCommand(exe string, args []string) Commander {
	c := &Command{
		cmd: execCommand(goCmd.Options{
			Streaming: true,
			Buffered:  false,
		}, exe, args...),
	}
	return c
}

func (c *Command) Status() goCmd.Status {
	return c.cmd.Status()
}

func (c *Command) Stop() error {
	return c.cmd.Stop()
}

func (c *Command) Streams() *Streams {
	return c.streams
}

// Runs the command, returning the status and the details/properties of the
// started process if it was successful.
func (c *Command) Start(env []string) (err error) {
	var startTimer <-chan time.Time

	c.cmd.Env = env

	c.streams = &Streams{
		Status: c.cmd.Start(),
		Stdout: c.cmd.Stdout,
		Stderr: c.cmd.Stderr,
	}

	startTimer = time.After(startTimerDuration)

	select {
	// Start timer
	case <-startTimer:
		// start timer expired, if exe is still up then it's probably fine
		startTimer = nil

		if !upCheck(c.Status().PID) {
			err = ErrTimeout{c.cmd.Name}
			return
		}

	case pStatus := <-c.streams.Status:
		// status from go-cmd
		if pStatus.Complete {
			if pStatus.Exit != 0 {
				err = ErrCmdExitedNonZero{pStatus.Exit}
			} else {
				err = ErrCmdExited{}
			}
			return
		} else {
			err = ErrUnhandledStatus{pStatus.Error.Error()}
		}
	}

	return
}

func isUp(pid int) bool {
	up := true

	p, err := os.FindProcess(pid)
	if err != nil {
		up = false
	} else {
		if err := p.Signal(syscall.Signal(0)); err != nil {
			up = false
		}
	}

	return up
}
