// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package cmd

import (
	"os"
	"reflect"
	"strconv"
	"testing"
	"time"

	goCmd "github.com/go-cmd/cmd"
)

func TestHelperExec(t *testing.T) {
	if os.Getenv("GO_WANT_HELPER_PROCESS") != "1" {
		return
	}

	ms, _ := strconv.Atoi(os.Getenv("RUNTIME_MS"))
	time.Sleep(time.Duration(ms) * time.Millisecond)

	ret, _ := strconv.Atoi(os.Getenv("EXIT_CODE"))
	os.Exit(ret)
}

func TestStart(t *testing.T) {
	subtests := []struct {
		name            string
		inRet           int
		inUpCheckRet    bool
		inRuntimeMs     int
		inStartTimerDur time.Duration
		expectedErr     error
	}{
		{
			name:            "started no failure",
			inUpCheckRet:    true,
			inRuntimeMs:     2000,
			inStartTimerDur: 100 * time.Millisecond,
			expectedErr:     nil,
		},
		{
			name:  "return non-zero",
			inRet: 4,
			// made really long so it doesn't 'normally' time out, which might
			// happen if it takes the OS/test system too long to actually run
			// the test exe
			inStartTimerDur: 5 * time.Second,
			expectedErr:     ErrCmdExitedNonZero{},
		},
		{
			name:  "return zero",
			inRet: 0,
			// made really long so it doesn't 'normally' time out, which might
			// happen if it takes the OS/test system too long to actually run
			// the test exe
			inStartTimerDur: 5 * time.Second,
			expectedErr:     ErrCmdExited{},
		},
		{
			name:            "timeout",
			inUpCheckRet:    false,
			inRuntimeMs:     2000,
			inStartTimerDur: 100 * time.Millisecond,
			expectedErr:     ErrTimeout{},
		},
	}

	var upCheckRet bool
	upCheck = func(pid int) bool {
		if pid == 0 {
			t.Fatalf("received unexpected pid: %d", pid)
		}

		return upCheckRet
	}
	defer func() {
		upCheck = isUp
		execCommand = goCmd.NewCmdOptions
		startTimerDuration = 100 * time.Millisecond
	}()

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			startTimerDuration = st.inStartTimerDur

			upCheckRet = st.inUpCheckRet

			execCommand = func(options goCmd.Options, name string, args ...string) *goCmd.Cmd {
				cmd := goCmd.NewCmdOptions(options, os.Args[0])
				cmd.Args = args
				return cmd
			}

			c := NewCommand("foo", []string{"-test.run=TestHelperExec"})
			env := []string{
				"GO_WANT_HELPER_PROCESS=1",
				"EXIT_CODE=" + strconv.Itoa(st.inRet),
				"RUNTIME_MS=" + strconv.Itoa(st.inRuntimeMs),
			}
			if err := c.Start(env); err != nil {
				// note: TypeOf used instead of errors.Is since an instance of
				// the error is stored in the subtest (expectedErr), and I'm
				// unaware of any way to "store a type" for the comparison here
				if reflect.TypeOf(err) != reflect.TypeOf(st.expectedErr) {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			} else {
				if st.expectedErr != nil {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			}
		})
	}
}
