// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"fmt"
	"reflect"
	"testing"
	"time"

	goCmd "github.com/go-cmd/cmd"
	"sr.ht/~craftyguy/superd/internal/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/pkg/api"
)

func helperService(name string) *Service {
	s := Init(name)
	s.config = &config.ServiceConfig{
		Name:            name,
		Type:            config.TypeSimple,
		Restart:         config.RestartNever,
		Cmd:             "foo",
		CmdArgs:         []string{"--bar"},
		StartLimitBurst: 5,
		StartLimitDur:   1 * time.Millisecond,
	}

	return s
}

type testCommand struct {
	cmd    string
	args   []string
	status chan goCmd.Status
}

func (c *testCommand) Status() goCmd.Status {
	return goCmd.Status{}
}

// this is defined by each test
var testCommandStartFunc func(c string, a []string) error

func (c *testCommand) Start(env []string) (err error) {
	return testCommandStartFunc(c.cmd, c.args)
}

func (c *testCommand) Stop() error {
	return nil
}

func (c *testCommand) Streams() *cmd.Streams {
	return &cmd.Streams{
		Status: c.status,
	}
}

func helperNewCommand(exe string, args []string) cmd.Commander {
	return &testCommand{
		cmd:  exe,
		args: args,
	}
}

func helperWaitForStatus(details chan api.ServiceDetails, status api.ServiceStatus) (err error) {

	start := time.Now()

	for d := range details {
		if d.Status == status {
			return
		}

		// can't use time.After here in the select because reading from
		// s.Status is a fire hose and continuously reading from it will
		// spin the loop too quickly and receiving from the time.After chan
		// won't happen (or won't happen for a loooong time)
		if time.Since(start) > 2*time.Second {
			return fmt.Errorf("timed out waiting for expected status")
		}
		time.Sleep(time.Millisecond * 1)
	}
	return
}

func TestServiceStatus(t *testing.T) {
	subtests := []struct {
		name      string
		inName    string
		expected  api.ServiceStatus
		startFunc func(c string, a []string) error
	}{
		{
			name:     "status starting",
			inName:   "testStatusStarting",
			expected: api.ServiceStarting,
			startFunc: func(c string, a []string) error {
				// sleep for some arbitrary time so that the "starting" status
				// has enough time to be received. if this is too short, it may
				// transistion to "started" before the status we want is sent
				time.Sleep(3 * time.Second)
				return nil
			},
		},
		{
			name:      "status started",
			inName:    "testStatusStarted",
			expected:  api.ServiceStarted,
			startFunc: func(c string, a []string) error { return nil },
		},
		{
			name:      "status failed",
			inName:    "testStatusFailed",
			expected:  api.ServiceFailed,
			startFunc: func(c string, a []string) error { return cmd.ErrCmdExitedNonZero{Ret: -1} },
		},
	}

	newCommand = helperNewCommand
	defer func() {
		newCommand = cmd.NewCommand
	}()

	logger = log.Default()

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			quit := make(chan struct{})

			s := helperService(subtest.inName)
			testCommandStartFunc = subtest.startFunc

			go s.Run(quit)

			select {
			case s.Start <- true:
			case <-time.After(1 * time.Second):
				t.Fatalf("timed out trying to start test process")
			}

			if err := helperWaitForStatus(s.Details, subtest.expected); err != nil {
				t.Fatal(err)
			}

			close(quit)
		})
	}
}

// testing the "stopped" status requires a bit more setup
func TestServiceStatusStopped(t *testing.T) {

	newCommand = helperNewCommand
	defer func() {
		newCommand = cmd.NewCommand
	}()

	logger = log.Default()

	quit := make(chan struct{})

	s := helperService("testServiceStatusStopped")
	testCommandStartFunc = func(cmd string, args []string) error { return nil }

	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	// wait for "started" status
	if err := helperWaitForStatus(s.Details, api.ServiceStarted); err != nil {
		t.Fatal(err)
	}

	// signal stop
	select {
	case s.Stop <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to stop test process")
	}

	// wait for "stopped" status
	if err := helperWaitForStatus(s.Details, api.ServiceStopped); err != nil {
		t.Fatal(err)
	}

	close(quit)
}

func TestExecOnFailure(t *testing.T) {

	newCommand = helperNewCommand
	defer func() {
		newCommand = cmd.NewCommand
	}()

	logger = log.Default()

	quit := make(chan struct{})

	startCmd := "startCmd"
	startArgs := []string{"--lime"}
	expectedCmd := "foo"
	expectedArgs := []string{"-t", "bar", "--bazz"}

	s := helperService("testExecOnFailure")
	s.config.ExecOnFailureCmd = expectedCmd
	s.config.ExecOnFailureCmdArgs = expectedArgs
	s.config.Cmd = startCmd
	s.config.CmdArgs = startArgs
	// the retry mechanism makes testing this more difficult, so disable it
	s.config.StartLimitBurst = 1

	expectedRun := make(chan struct{})
	testCommandStartFunc = func(c string, a []string) error {
		if c == startCmd {
			// override start func to test when ExecOnFailure is run
			testCommandStartFunc = func(c string, a []string) error {
				if c != expectedCmd {
					t.Fatalf("expected: %s, got: %s\n", expectedCmd, c)
				} else if !reflect.DeepEqual(a, expectedArgs) {
					t.Fatalf("expected: %q, got: %q\n", expectedArgs, a)
				} else {
					close(expectedRun)
				}
				return cmd.ErrCmdExited{}
			}

			// fake the service failure
			return cmd.ErrCmdExitedNonZero{Ret: -1}
		} else {
			t.Fatalf("unexpected command run by service: %s", c)
		}
		return nil
	}

	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	select {
	case <-expectedRun:
	case <-time.After(2 * time.Second):
		t.Fatal("timed out waiting for ExecOnFailure to run")
	}

	close(quit)
}

func TestExecStartPre(t *testing.T) {

	newCommand = helperNewCommand
	defer func() {
		newCommand = cmd.NewCommand
	}()

	logger = log.Default()

	quit := make(chan struct{})

	expectedCmd := "do_stuff_pre"
	expectedArgs := []string{"-a", "thing", "--mango"}

	s := helperService("testExecStartPre")
	s.config.ExecStartPreCmd = expectedCmd
	s.config.ExecStartPreCmdArgs = expectedArgs

	// chan closed when event is run
	expectedRun := make(chan struct{})
	testCommandStartFunc = func(c string, a []string) error {
		if c == expectedCmd && reflect.DeepEqual(a, expectedArgs) {
			close(expectedRun)
		}
		return nil
	}

	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	select {
	case <-expectedRun:
	case <-time.After(2 * time.Second):
		t.Fatal("timed out waiting for ExecStartPre to run")
	}

	close(quit)
}

func TestFailureRestart(t *testing.T) {

	newCommand = helperNewCommand
	defer func() {
		newCommand = cmd.NewCommand
	}()

	logger = log.Default()

	quit := make(chan struct{})

	s := helperService("testFailureRestart")
	s.config.Restart = config.RestartFailure

	expectedRun := make(chan struct{})
	totalRuns := 0

	testCommandStartFunc = func(c string, a []string) error {
		// fake the service failure
		totalRuns++
		if totalRuns == s.config.StartLimitBurst {
			t.Log(totalRuns)
			close(expectedRun)
		}
		return cmd.ErrCmdExitedNonZero{Ret: -1}
	}

	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	select {
	case <-expectedRun:
	case <-time.After(2 * time.Second):
		t.Fatal("timed out waiting for ExecStartPre to run")
	}

	close(quit)
}

func TestShouldRestart(t *testing.T) {
	subtests := []struct {
		name      string
		inStatus  api.ServiceStatus
		inType    config.Type
		inRestart config.RestartCond
		expected  bool
	}{
		{
			name:      "simple failed always restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "simple stopped always restart",
			inStatus:  api.ServiceStopped,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "simple failed never restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartNever,
			expected:  false,
		},
		{
			name:      "simple failed restart on failure",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartFailure,
			expected:  true,
		},
		{
			name:      "simple starting always restart",
			inStatus:  api.ServiceStarting,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot stopped always restart",
			inStatus:  api.ServiceStopped,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot failed always restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "oneshot started always restart",
			inStatus:  api.ServiceStarted,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot started restart on failure",
			inStatus:  api.ServiceStarted,
			inType:    config.TypeOneshot,
			inRestart: config.RestartFailure,
			expected:  false,
		},
	}

	logger = log.Default()

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			s := helperService(st.name)
			s.config.Type = st.inType
			s.config.Restart = st.inRestart

			var out bool
			switch st.inType {
			case config.TypeSimple:
				out = s.simpleShouldRestart(st.inStatus)
			case config.TypeOneshot:
				out = s.oneshotShouldRestart(st.inStatus)
			default:
				t.Fatalf("unexpected type %s", st.inType)
			}
			if out != st.expected {
				t.Fatalf("expected: %t, received: %t", st.expected, out)
			}
		})
	}
}

func TestBeforeAfter(t *testing.T) {
	subtests := []struct {
		name string
		in   []string
	}{
		{
			name: "before / after",
			in:   []string{"foo", "banana", "something"},
		},
	}

	for _, st := range subtests {

		t.Run(st.name, func(t *testing.T) {
			s := helperService("test")
			for _, i := range st.in {
				s.AddAfter(i)
				s.AddBefore(i)
			}

			out := s.After()
			if !reflect.DeepEqual(out, st.in) {
				t.Fatalf("Expected: %q, got: %q", st.in, out)
			}

			out = s.Before()
			if !reflect.DeepEqual(out, st.in) {
				t.Fatalf("Expected: %q, got: %q", st.in, out)
			}
		})
	}
}
